import aiohttp
import asyncio
import pandas as pd
import requests
import nest_asyncio
import logging

from datetime import datetime


class FredDownloader(object):
    def __init__(self, api_key: str):
        self.__api_key = api_key
        self.__uri = 'https://api.stlouisfed.org/fred'
        self.__headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3'
        }
        self.__nan_char = '.'
        
        nest_asyncio.apply()

    @staticmethod
    def __compare_date(date_1: str, date_2: str, compare_type: str):
        if compare_type == 'min':
            return datetime.strftime(min(datetime.strptime(date_1, '%Y-%m-%d'), datetime.strptime(date_2, '%Y-%m-%d')), '%Y-%m-%d')
        
        elif compare_type == 'max':
            return datetime.strftime(max(datetime.strptime(date_1, '%Y-%m-%d'), datetime.strptime(date_2, '%Y-%m-%d')), '%Y-%m-%d')
        
        else:
            raise ValueError('Invalid compare type')
    
    def __convert_to_float(self, value: str):
        if value == self.__nan_char:
            return float('nan')

        else:
            return float(value.replace(',', ''))
        
    def search_series_id_by_keyword(self, keyword: str):
        url = f'{self.__uri}/series/search'
        resp = requests.get(url=url, params={'api_key': self.__api_key, 'search_text': keyword, 'file_type': 'json', 'order_by': 'search_rank'}).json()
        
        try:
            data = resp['seriess']

            return pd.DataFrame(data)[['title', 'id', 'observation_start', 'observation_end', 'frequency_short', 'units_short', 'seasonal_adjustment_short']]

        except KeyError:
            logging.error('Error in searching keyword')
    
    async def __fetch_data(self, session, url, params):
        async with session.get(url, params=params, headers=self.__headers) as response:
            return await response.json()
    
    async def __get_series_info(self, session, series_id):
        url = f'{self.__uri}/series'
        resp = await self.__fetch_data(session, url, params={'api_key': self.__api_key, 'series_id': series_id, 'file_type': 'json'})

        return resp
    
    async def __get_series(self, session, series_id, start, end):
        series_info = await self.__get_series_info(session, series_id)

        try:
            adjusted_start = self.__compare_date(series_info['seriess'][0]['observation_start'], start, 'max')
            adjusted_end = self.__compare_date(series_info['seriess'][0]['observation_end'], end, 'min')
            url = f'{self.__uri}/series/observations'

            resp = await self.__fetch_data(session, url, params={'api_key': self.__api_key, 'series_id': series_id, 'observation_start': adjusted_start, 'observation_end': adjusted_end, 'file_type': 'json'})
            resp_dict = [{'date': observation['date'], 'value': self.__convert_to_float(observation['value'])} for observation in resp['observations']]

            df = pd.DataFrame(resp_dict)
            df.set_index('date', inplace=True)
            df['value'] = df['value'].astype(float)
            df.index = pd.to_datetime(df.index)
            df.sort_index(inplace=True)

            return series_id, df
        
        except KeyError:
            logging.error(f'Invalid series id: {series_id}')

            return series_id, None

    async def __get_multiple_series(self, series_ids, start, end):
        async with aiohttp.ClientSession() as session:
            tasks = [self.__get_series(session, series_id, start, end) for series_id in series_ids]

            return dict(await asyncio.gather(*tasks))

    def download_data(self, series_ids, start, end):
        return asyncio.run(self.__get_multiple_series(series_ids, start, end))



import aiohttp
import asyncio
import requests
import re
import pandas as pd
import nest_asyncio
import logging
import re

from typing import List, Tuple


class BokDownloader(object):
    AVAILABLE_PERIOD = ['A', 'Q', 'M', 'D']
    VALID_FORMATS = {'A': r'^\d{4}$', 'Q':  r'^\d{4}Q[1-4]$', 'M': r'^\d{4}(0[1-9]|1[0-2])$', 'D':  r'^\d{4}(0[1-9]|1[0-2])(0[1-9]|[12][0-9]|3[01])$'}

    def __init__(self, api_key):
        self.__api_key = api_key
        self.__uri = 'https://ecos.bok.or.kr/api'
        self.__headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3'
        }

        self.available_stats = self.__get_stat_table_list()

        nest_asyncio.apply()
    
    def __check_format(self, period, input_date):
        if re.match(BokDownloader.VALID_FORMATS[period], input_date):
            return True
        
        else:
            return False
    
    @staticmethod
    def __convert_to_datetime(period, index):
        if period == 'A':
            return pd.to_datetime(index,  format='%Y') + pd.offsets.YearEnd(0)
        
        elif period == 'Q':
            pd.to_datetime(index.apply(lambda x: pd.Period(x, freq='Q').start_time)) + pd.offsets.QuarterEnd()

        elif period == 'M':
            return pd.to_datetime(index, format='%Y%m') + pd.DateOffset(day=1)
        
        elif period == 'D':
            return pd.to_datetime(index, format='%Y%m%d')

        else:
            raise ValueError('Invalid period')
    
    def __get_stat_table_list(self):
        url = f'{self.__uri}/StatisticTableList/{self.__api_key}/json/kr/1/99999/'

        try:
            response = requests.get(url=url).json()
            datas = response['StatisticTableList']['row']
            
            for data in datas:
                data['STAT_NAME'] = re.sub(r'\d+\.', '', data['STAT_NAME'])
            
            df = pd.DataFrame(datas)

            return df
        
        except KeyError:
            logging.error('Error in fetching table.')

    def search_stat_code_by_keyword(self, keyword: str):
        return self.available_stats.query(f"STAT_NAME.str.contains('{keyword}')")
    
    def __make_code(self, stat_code, item_codes):
        parsed_item_codes = [code for code in item_codes if code != '']
        code_str = '-'.join(parsed_item_codes)

        return f'{stat_code}-{code_str}'

    async def __fetch_data(self, session, url):
        async with session.get(url, headers=self.__headers) as response:
            return await response.json()
    
    async def __get_series(self, session, stat_code, period, start, end, item_code_1='', item_code_2='', item_code_3='', item_code_4=''):
        if period not in BokDownloader.AVAILABLE_PERIOD:
            raise ValueError('Invalid period')
        
        elif (self.__check_format(period, start) or self.__check_format(period, end)) is False:
            raise ValueError('Invalid date format')

        url = f'{self.__uri}/StatisticSearch/{self.__api_key}/json/en/1/10000/{stat_code}/{period}/{start}/{end}/{item_code_1}/{item_code_2}/{item_code_3}/{item_code_4}'
        raw_datas = await self.__fetch_data(session, url)
        code = self.__make_code(stat_code, [item_code_1, item_code_2, item_code_3, item_code_4])

        try:
            data = [{'time': raw_data['TIME'], 'value': float(raw_data['DATA_VALUE'].replace(',', ''))} for raw_data in raw_datas['StatisticSearch']['row']]
            df = pd.DataFrame(data)
            df['value'] = df['value'].astype(float)
            df.set_index('time', inplace=True)
            df.index = self.__convert_to_datetime(period, df.index)

            return f'{stat_code}-{code}', df

        except KeyError:
            logging.error(f'Invalid series id: {stat_code}-{code}')

            return f'{stat_code}-{code}', None
    
    async def __get_multiple_series(self, request_infos: List[Tuple[str]]):
        async with aiohttp.ClientSession() as session:
            tasks = [self.__get_series(session, stat_code=request_info[0], period=request_info[1], start=request_info[2], end=request_info[3], 
                                       item_code_1=request_info[4], item_code_2=request_info[5], item_code_3=request_info[6], item_code_4=request_info[7]) for request_info in request_infos]

            return dict(await asyncio.gather(*tasks))
        
    def download_data(self, request_infos: List[Tuple[str]]):
        return asyncio.run(self.__get_multiple_series(request_infos))




from econdatareader.version import version as __version__
from econdatareader.fred import FredDownloader
from econdatareader.finance import FinanceDownloader
from econdatareader.bok import BokDownloader



from setuptools import setup, find_packages

with open('readme.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='econ-datareader',
    version='1.0.5',
    description='Download Econ Data - Macro and Finance',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='WyChoi1995',
    author_email='wydanielchoi@gmail.com',
    url='https://github.com/WYChoi1995/econ-datareader',
    install_requires=['pandas', 'aiohttp', 'nest_asyncio',],
    packages=find_packages(exclude=[]),
    keywords=['finance', 'economics'],
    python_requires='>=3.6',
    package_data={},
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)


import aiohttp
import asyncio
import logging
import pandas as pd
import re
import nest_asyncio

from datetime import datetime

class FinanceDownloader(object):
    KOREA_STOCK_VALID_INTERVAL = ['minute', 'minute3', 'minute5', 'minute10', 'minute30', 'minute60', 'day', 'week', 'month']
    GLOBAL_FINANCE_DATA_VALID_INTERVAL = ['1m', '2m', '5m', '15m', '30m', '60m', '90m', '1h', '1d']
    CRYPTO_BINANCE_VALID_INTERVAL = ['1s', '1m', '3m', '5m', '15m', '30m', '1h', '2h', '4h', '6h', '8h', '12h', '1d', '3d', '1w', '1M']
    CRYTPO_UPBIT_VALID_INTERVAL = ['1m', '3m', '5m', '10m', '15m', '30m', '60m', '240m', '1d', '1w', '1M']
    TIME_FACTORS = {'s': 1000, 'm': 60 * 1000, 'h': 3600 * 1000, 'd': 86400 * 1000, 'w': 604800 * 1000, 'M': 2592000 * 1000}
    GLOBAL_CANDLE_LIMITS = {'m': 7 * 24 * 60 * 60, 'h': 30 * 24 * 60 * 60, 'd': 365 * 24 * 60 * 60}
    AVAILABLE_DOWNLOAD_TYPE = ['KOREA_STOCK', 'GLOBAL_FINANCE', 'CRYPTO_SPOT_BINANCE', 'CRYPTO_FUTURES_BINANCE', 'CRYPTO_SPOT_UPBIT']

    def __init__(self):
        self.__binance_spot_uri = 'https://api.binance.com/api/v3/klines'
        self.__binance_futures_uri = 'https://fapi.binance.com/fapi/v1/klines'
        self.__naver_finance_uri = 'https://api.stock.naver.com/chart/domestic/item/'
        self.__yahoo_finance_uri = 'https://query1.finance.yahoo.com/v8/finance/chart/'
        self.__upbit_spot_uri = 'https://api.upbit.com/v1/candles/'
        self.__headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3'
        }
        self.__binance_rate_count = 0
        self.__upbit_rate_count_sec = 9
        self.__upbit_rate_count_min = 600

        nest_asyncio.apply()
    
    @staticmethod
    def __convert_interval_to_ms(interval: str):
        return int(interval[:-1]) * FinanceDownloader.TIME_FACTORS[interval[-1]]
    
    @staticmethod
    def __convert_interval_to_sec(interval: str):
        return int(interval[:-1]) * FinanceDownloader.GLOBAL_CANDLE_LIMITS[interval[-1]]

    @staticmethod
    def __convert_datetime_to_ts(date: int, is_milli: bool = False):
        datetime_obj = datetime.strptime(str(date), '%Y%m%d%H%M')

        if is_milli:
            return 1000 * int(datetime_obj.timestamp())
        
        else:
            return int(datetime_obj.timestamp())
    
    @staticmethod
    def __convert_datetime_to_iso8061(date: int):
        str_date = str(date)
        return f'{str_date[:4]}-{str_date[4:6]}-{str_date[6:8]}T{str_date[8:10]}:{str_date[10:12]}:00Z'
    
    @staticmethod
    def __determine_interval(interval: str):
        if 'm' in interval:
            return f'minutes/{interval[:-1]}'

        elif 'd' in interval:
            return 'days'
        
        elif 'w' in interval:
            return 'weeks'
        
        elif 'M' in interval:
            return 'months'
        
        else:
            raise ValueError(f'Unsupported interval {interval}')
    
    @staticmethod
    def __parse_time_values(input_string: str):
        min_match = re.search(r'min=(\d+)', input_string)
        sec_match = re.search(r'sec=(\d+)', input_string)

        if min_match and sec_match:
            min_value = int(min_match.group(1))
            sec_value = int(sec_match.group(1))

            return min_value, sec_value
        
        else:
            return None, None
    
    async def __fetch_data_crypto(self, session, url, params, is_binance=True):
        async with session.get(url, params=params) as response:
            if is_binance:
                self.__binance_rate_count = float(response.headers.get('X-MBX-USED-WEIGHT-1M'))

            else:
                self.__upbit_rate_count_min, self.__upbit_rate_count_sec = self.__parse_time_values(response.headers.get('Remaining-Req'))
                
            return await response.json()
    
    async def __fetch_data_tradfi(self, session, url, params):
        async with session.get(url, params=params, headers=self.__headers) as response:
            return await response.json()
    
    async def __get_korea_stock_data(self, session, ticker, interval, start, end):
        url = f'{self.__naver_finance_uri}{ticker}/{interval}'

        resp = await self.__fetch_data_tradfi(session, url, params={'startDateTime': start, 'endDateTime': end})

        if 'code' not in resp:
            if 'minute' in interval:
                df = pd.DataFrame(resp)
                df.drop_duplicates(inplace=True, subset=['localDateTime'])
                df.set_index('localDateTime', inplace=True)
                df.index = pd.to_datetime(df.index, format='%Y%m%d%H%M%S')
                df.sort_index(inplace=True)
                df.ffill(inplace=True)
                
                return ticker, df
            
            else:
                df = pd.DataFrame(resp)
                df.drop_duplicates(inplace=True, subset=['localDate'])
                df.set_index('localDate', inplace=True)
                df.index = pd.to_datetime(df.index, format='%Y%m%d')
                df.sort_index(inplace=True)
                df.ffill(inplace=True)
                
                return ticker, df
        
        else:
            logging.error(f'Error on fetching: Code {resp["code"]} Msg {resp["message"]}')
            return
    
    async def __get_global_finance_data(self, session, ticker, interval, start, end):
        url = f'{self.__yahoo_finance_uri}{ticker}'
        datas = []
        interval_convert = self.__convert_interval_to_sec(interval)
        end_time = self.__convert_datetime_to_ts(end)
        last_end_time = self.__convert_datetime_to_ts(start)
        
        while end_time > last_end_time:
            start_time = end_time - interval_convert
            resp  = await self.__fetch_data_tradfi(session, url, params={'interval': interval, 'period1': start_time, 'period2': end_time})

            if resp['chart']['result'] is not None:
                try:
                    ohlcv = resp['chart']['result'][0]['indicators']['quote'][0]
                    adj_close = resp['chart']['result'][0]['indicators']['adjclose'][0]['adjclose']
                    
                    data = {
                                'timestamp': resp['chart']['result'][0]['timestamp'],
                                'open': ohlcv['open'],
                                'high': ohlcv['high'],
                                'low': ohlcv['low'],
                                'close': ohlcv['close'],
                                'volume': ohlcv['volume'],
                                'adj_close': adj_close
                            }
                    
                    datas.append(data)

                except KeyError:
                    ohlcv = resp['chart']['result'][0]['indicators']['quote'][0]
                    
                    data = {
                                'timestamp': resp['chart']['result'][0]['timestamp'],
                                'open': ohlcv['open'],
                                'high': ohlcv['high'],
                                'low': ohlcv['low'],
                                'close': ohlcv['close'],
                                'volume': ohlcv['volume']
                            }
                    
                    datas.append(data)

                end_time = start_time
            
            else:
                logging.error('Fetcherror: {}'.format(resp['chart']['error']))
                break
        
        df = pd.concat([pd.DataFrame(data) for data in datas], axis=0)
        df.drop_duplicates(inplace=True, subset=['timestamp'])
        df.set_index('timestamp', inplace=True)
        df.index = pd.to_datetime(df.index, unit='s')
        df.sort_index(inplace=True)
        df.ffill(inplace=True)
        
        return ticker, df

    async def __get_binance_futures_data(self, session, ticker, interval, start, end):
        candle_datas = []
        interval_convert = self.__convert_interval_to_ms(interval)
        start_time = self.__convert_datetime_to_ts(start, True)
        end_time = self.__convert_datetime_to_ts(end, True)
        temp_start = end_time - 499 * interval_convert

        if temp_start <= start_time:
            params = {'symbol': ticker, 'interval': interval, 'startTime': start_time, 'endTime': end_time}
            resp = await self.__fetch_data_crypto(session, self.__binance_futures_uri, params)
            candle_datas.append(resp)

        else:
            while temp_start > start_time:
                if self.__binance_rate_count >= 1500:
                    await asyncio.sleep(60)
                    self.__binance_rate_count = 0

                params = {'symbol': ticker, 'interval': interval, 'startTime': temp_start, 'endTime': end_time}
                resp = await self.__fetch_data_crypto(session, self.__binance_futures_uri, params)

                if 'code' not in resp:
                    candle_datas.append(resp)

                else:
                    logging.error(f'{resp}')
                    break

                end_time = temp_start
                temp_start = end_time - 499 * interval_convert
    
        df = pd.DataFrame([candle_datum for candle_data in candle_datas for candle_datum in candle_data])[[0,1,2,3,4,5,9]]
        df.columns = ['time', 'open', 'high', 'low', 'close', 'volume', 'takerBuyBase']
        df['open'] = df['open'].astype(float)
        df['high'] = df['high'].astype(float)
        df['low'] = df['low'].astype(float)
        df['close'] = df['close'].astype(float)
        df['volume'] = df['volume'].astype(float)
        df['takerBuyBase'] = df['takerBuyBase'].astype(float)
        df.drop_duplicates(inplace=True, subset=['time'])
        df.set_index('time', inplace=True)
        df.index = pd.to_datetime(df.index, unit='ms')
        df.sort_index(inplace=True)

        return ticker, df
    
    async def __get_binance_spot_data(self, session, ticker, interval ,start, end):
        candle_datas = []
        interval_convert = self.__convert_interval_to_ms(interval)
        start_time = self.__convert_datetime_to_ts(start, True)
        end_time = self.__convert_datetime_to_ts(end, True)
        temp_start = end_time - 500 * interval_convert

        if temp_start <= start_time:
            params = {'symbol': ticker, 'interval': interval, 'startTime': start_time, 'endTime': end_time}
            resp = await self.__fetch_data_crypto(session, self.__binance_spot_uri, params)
            candle_datas.append(resp)

        else:
            while temp_start > start_time:
                if self.__binance_rate_count >= 1500:
                    await asyncio.sleep(60)
                    self.__binance_rate_count = 0

                params = {'symbol': ticker, 'interval': interval, 'startTime': temp_start, 'endTime': end_time}
                resp = await self.__fetch_data_crypto(session, self.__binance_spot_uri, params)

                if 'code' not in resp:   
                    candle_datas.append(resp)

                else:
                    logging.error(f'{resp}')
                    break
                
                end_time = temp_start
                temp_start = end_time - 500 * interval_convert
        try:
            df = pd.DataFrame([candle_datum for candle_data in candle_datas for candle_datum in candle_data])[[0,1,2,3,4,5,9]]
            df.columns = ['time', 'open', 'high', 'low', 'close', 'volume', 'takerBuyBase']
            df['open'] = df['open'].astype(float)
            df['high'] = df['high'].astype(float)
            df['low'] = df['low'].astype(float)
            df['close'] = df['close'].astype(float)
            df['volume'] = df['volume'].astype(float)
            df['takerBuyBase'] = df['takerBuyBase'].astype(float)
            df.drop_duplicates(inplace=True, subset=['time'])
            df.set_index('time', inplace=True)
            df.index = pd.to_datetime(df.index, unit='ms')
            df.sort_index(inplace=True)
            
            return ticker, df

        except KeyError:
            logging.error(f'{ticker}: Unable to make dataframe')

            return ticker, None

    async def __get_upbit_spot_data(self, session, ticker, interval, start, end):
        candle_datas = []
        end_time_param = self.__convert_datetime_to_iso8061(end)
        start_time = self.__convert_datetime_to_ts(start, True)
        end_time = self.__convert_datetime_to_ts(end, True)
        url = f'{self.__upbit_spot_uri}{self.__determine_interval(interval)}'

        while end_time > start_time:
            if self.__upbit_rate_count_min < 10:
                await asyncio.sleep(60)
                self.__upbit_rate_count_min = 600
            
            elif self.__upbit_rate_count_sec < 2:
                await asyncio.sleep(1)
                self.__upbit_rate_count_sec = 9

            else:
                params = {'market': ticker, 'count': 200, 'to': end_time_param}
                resp = await self.__fetch_data_crypto(session, url, params, False)

                candle_datas.append(resp)
                end_time_param = resp[-1]['candle_date_time_utc']
                end_time = resp[-1]['timestamp']
        
        try:
            df = pd.concat([pd.DataFrame(data) for data in candle_datas], axis=0)
            df.columns = ['market', 'time', 'time_KST', 'open', 'high', 'low', 'close', 'ts', 'amt', 'volume', 'unit']
            df.drop(columns=['market', 'time_KST', 'unit', 'ts'], inplace=True)
            df.drop_duplicates(inplace=True, subset=['time'])
            df.set_index('time', inplace=True)
            df.index = pd.to_datetime(df.index)
            df.sort_index(inplace=True)
        
            return ticker, df
        
        except KeyError:
            logging.error(f'{ticker}: Unable to make dataframe')

            return ticker, None

    async def __get_multiple_korea_stock_data(self, tickers, interval, start, end):
        if interval not in FinanceDownloader.KOREA_STOCK_VALID_INTERVAL:
            raise ValueError(f'Invalid interval, Valid interval is {FinanceDownloader.KOREA_STOCK_VALID_INTERVAL}')

        async with aiohttp.ClientSession() as session:
            tasks = [self.__get_korea_stock_data(session, ticker, interval, start, end) for ticker in tickers]

            return dict(await asyncio.gather(*tasks))
    
    async def __get_multiple_global_finance_data(self, tickers,  interval, start, end):
        if interval not in FinanceDownloader.GLOBAL_FINANCE_DATA_VALID_INTERVAL:
            raise ValueError(f'Invalid interval, Valid interval is {FinanceDownloader.GLOBAL_FINANCE_DATA_VALID_INTERVAL}')
        
        async with aiohttp.ClientSession() as session:
            tasks = [self.__get_global_finance_data(session, ticker, interval, start, end) for ticker in tickers]

            return dict(await asyncio.gather(*tasks))

    async def __get_multiple_binance_spot_data(self, tickers, interval, start, end):
        if interval not in FinanceDownloader.CRYPTO_BINANCE_VALID_INTERVAL:
            raise ValueError(f'Invalid interval, Valid interval is {FinanceDownloader.CRYPTO_BINANCE_VALID_INTERVAL}')

        async with aiohttp.ClientSession() as session:
            tasks = [self.__get_binance_spot_data(session, ticker, interval, start, end) for ticker in tickers]

            return dict(await asyncio.gather(*tasks))

    async def __get_multiple_binance_futures_data(self, tickers, interval, start, end):

        if interval not in FinanceDownloader.CRYPTO_BINANCE_VALID_INTERVAL:
            raise ValueError(f'Invalid interval, Valid interval is {FinanceDownloader.CRYPTO_BINANCE_VALID_INTERVAL}')
        
        async with aiohttp.ClientSession() as session:
            tasks = [self.__get_binance_futures_data(session, ticker, interval, start, end) for ticker in tickers]

            return dict(await asyncio.gather(*tasks))
    
    async def __get_multiple_upbit_spot_data(self, tickers, interval, start, end):
        if interval not in FinanceDownloader.CRYTPO_UPBIT_VALID_INTERVAL:
            raise ValueError(f'Invalid interval, Valid interval is {FinanceDownloader.CRYTPO_UPBIT_VALID_INTERVAL}')
        
        async with aiohttp.ClientSession() as session:
            tasks = [self.__get_upbit_spot_data(session, ticker, interval, start, end) for ticker in tickers]

            return dict(await asyncio.gather(*tasks))
    
    def download_data(self, download_type, tickers, interval, start, end):
        '''
        Datetime Format: YYYYMMDDHHMM
        '''
        if download_type not in FinanceDownloader.AVAILABLE_DOWNLOAD_TYPE:
            raise ValueError(f'Invalid download type, Valid type is {FinanceDownloader.AVAILABLE_DOWNLOAD_TYPE}')
        
        else:
            if download_type == 'KOREA_STOCK':
                return asyncio.run(self.__get_multiple_korea_stock_data(tickers, interval, start, end))
            
            elif download_type == 'GLOBAL_FINANCE':
                return asyncio.run(self.__get_multiple_global_finance_data(tickers, interval, start, end))
            
            elif download_type == 'CRYPTO_SPOT_BINANCE':
                return asyncio.run(self.__get_multiple_binance_spot_data(tickers, interval, start, end))
            
            elif download_type == 'CRYPTO_FUTURES_BINANCE':
                return asyncio.run(self.__get_multiple_binance_futures_data(tickers, interval, start, end))
            
            elif download_type == 'CRYPTO_SPOT_UPBIT':
                return asyncio.run(self.__get_multiple_upbit_spot_data(tickers, interval, start, end))
